@@ -7,6 +7,10 @@ from torch.autograd import Function
 from functools import partial, reduce, wraps
 from itertools import chain
 from operator import mul
+
+from local_attention import LocalAttention
+from axial_positional_embedding import AxialPositionalEmbedding
+from product_key_memory import PKM
 from reformer_pytorch.reversible import ReversibleSequence
 
 #constants
@@ -42,6 +46,9 @@ def chunked_sum(tensor, chunks=1):
 def default(val, default_val):
     return default_val if val is None else val
 
+def cast_tuple(x):
+    return x if isinstance(x, tuple) else (x,)
+
 def max_neg_value(tensor):
     return -torch.finfo(tensor.dtype).max
 
@@ -76,13 +83,6 @@ def cache_method_decorator(cache_attr, cache_namespace, reexecute = False):
         return wrapper
     return inner_fn
 
-def look_around(x, backward = 1, forward = 0, pad_value = -1, dim = 2):
-    t = x.shape[1]
-    dims = (len(x.shape) - dim) * (0, 0)
-    padded_x = F.pad(x, (*dims, backward, forward), value= pad_value)
-    tensors = [padded_x[:, ind:(ind + t), ...] for ind in range(forward + backward + 1)]
-    return torch.cat(tensors, dim=dim)
-
 def expand_dim(dim, k, t):
     t = t.unsqueeze(dim)
     expand_shape = [-1] * len(t.shape)
@@ -295,7 +295,7 @@ class LSHAttention(nn.Module):
         # attention softmax, but normalizing keys is needed so that similarity for
         # the purposes of attention correctly corresponds to hash locality.
         bq = bqk
-        bk = F.normalize(bqk, p=2, dim=-1).type(bq.type())
+        bk = F.normalize(bqk, p=2, dim=-1).type_as(bq)
 
         # Allow each chunk to attend within itself, and also one chunk back. Chunk
         # boundaries might occur in the middle of a sequence of items from the
@@ -387,7 +387,7 @@ class LSHAttention(nn.Module):
 
         # Softmax.
         dots_logsumexp = torch.logsumexp(dots, dim=-1, keepdim=True)
-        dots = torch.exp(dots - dots_logsumexp).type(dots.type())
+        dots = torch.exp(dots - dots_logsumexp).type_as(dots)
         dropped_dots = self.dropout(dots)
 
         bo = torch.einsum('buij,buje->buie', dropped_dots, bv)
@@ -435,75 +435,6 @@ class LSHAttention(nn.Module):
         # return output, attention matrix, and bucket distribution
         return out, attn, buckets
 
-# local attention
-
-class LocalAttention(nn.Module):
-    def __init__(self, bucket_size, causal = False, look_backward = 1, look_forward = 0, dropout = 0., shared_qk = False):
-        super().__init__()
-        assert not (causal and look_forward > 0), 'you cannot look forward if causal'
-        self.bucket_size = bucket_size
-        self.causal = causal
-        self.look_backward = look_backward
-        self.look_forward = look_forward
-        self.shared_qk = shared_qk
-        self.dropout = nn.Dropout(dropout)
-
-    def forward(self, q, k, v, input_mask = None):
-        b, t, e, device, dtype = *q.shape, q.device, q.dtype
-        bucket_size, causal, look_backward, look_forward, shared_qk = self.bucket_size, self.causal, self.look_backward, self.look_forward, self.shared_qk
-
-        buckets = t // bucket_size
-
-        if shared_qk:
-            k = F.normalize(k, 2, dim=-1).type(q.type())
-
-        ticker = torch.arange(t, device=device, dtype=dtype)[None, :]
-        b_t = ticker.reshape(1, buckets, bucket_size)
-
-        bucket_fn = lambda t: t.reshape(b, buckets, bucket_size, -1)
-        bq, bk, bv = map(bucket_fn, (q, k, v))
-
-        look_around_kwargs = {'backward': look_backward, 'forward': look_forward}
-        bk = look_around(bk, **look_around_kwargs)
-        bv = look_around(bv, **look_around_kwargs)
-
-        bq_t = b_t
-        bq_k = look_around(b_t, **look_around_kwargs)
-
-        dots = torch.einsum('bhie,bhje->bhij', bq, bk) * (e ** -0.5)
-        mask_value = max_neg_value(dots)
-
-        if shared_qk:
-            mask = bq_t[:, :, :, None] == bq_k[:, :, None, :]
-            dots.masked_fill_(mask, TOKEN_SELF_ATTN_VALUE)
-            del mask
-
-        if causal:
-            mask = bq_t[:, :, :, None] < bq_k[:, :, None, :]
-            dots.masked_fill_(mask, mask_value)
-            del mask
-
-        mask = bq_k[:, :, None, :] == -1
-        dots.masked_fill_(mask, mask_value)
-        del mask
-
-        if input_mask is not None:
-            h = b // input_mask.shape[0]
-            input_mask = input_mask.reshape(-1, buckets, bucket_size)
-            mq = mk = input_mask
-            mk = look_around(mk, pad_value=False, **look_around_kwargs)
-            mask = (mq[:, None, :, :, None] * mk[:, None, :, None, :])
-            mask = merge_dims(0, 1, mask.expand(-1, h, -1, -1, -1))
-            dots.masked_fill_(~mask, mask_value)
-            del mask
-
-        attn = dots.softmax(dim=-1)
-        attn = self.dropout(attn)
-
-        out = torch.einsum('bhij,bhje->bhie', attn, bv)
-        out = out.reshape(b, t, e)
-        return out
-
 # simple full attention
 
 class FullQKAttention(nn.Module):
@@ -518,7 +449,7 @@ class FullQKAttention(nn.Module):
         t = query_len
 
         q = qk[:, 0:query_len]
-        qk = F.normalize(qk, 2, dim=-1).type(q.type())
+        qk = F.normalize(qk, 2, dim=-1).type_as(q)
 
         dot = torch.einsum('bie,bje->bij', q, qk) * (dim ** -0.5)
 
@@ -552,21 +483,25 @@ class FullQKAttention(nn.Module):
 # Shared qk attention, using either full or LSH attention
 
 class LSHSelfAttention(nn.Module):
-    def __init__(self, dim, heads = 8, bucket_size = 64, n_hashes = 8, causal = False, attn_chunks = 1, random_rotations_per_head = False, attend_across_buckets = True, allow_duplicate_attention = True, num_mem_kv = 0, one_value_head = False, use_full_attn = False, full_attn_thres = None, return_attn = False, post_attn_dropout = 0., dropout = 0., n_local_attn_heads = 0, **kwargs):
+    def __init__(self, dim, heads = 8, bucket_size = 64, n_hashes = 8, causal = False, dim_head = None, attn_chunks = 1, random_rotations_per_head = False, attend_across_buckets = True, allow_duplicate_attention = True, num_mem_kv = 0, one_value_head = False, use_full_attn = False, full_attn_thres = None, return_attn = False, post_attn_dropout = 0., dropout = 0., n_local_attn_heads = 0, **kwargs):
         super().__init__()
         assert dim % heads == 0, 'dimensions must be divisible by number of heads'
         assert n_local_attn_heads < heads, 'local attention heads must be less than number of heads'
 
+        dim_head = default(dim_head, dim // heads)
+        dim_heads = dim_head * heads
+
         self.dim = dim
         self.heads = heads
+        self.dim_head = dim_head
         self.attn_chunks = default(attn_chunks, 1)
 
         self.v_head_repeats = (heads if one_value_head else 1)
-        v_dim = dim // self.v_head_repeats
+        v_dim = dim_heads // self.v_head_repeats
 
-        self.toqk = nn.Linear(dim, dim, bias = False)
+        self.toqk = nn.Linear(dim, dim_heads, bias = False)
         self.tov = nn.Linear(dim, v_dim, bias = False)
-        self.to_out = nn.Linear(dim, dim)
+        self.to_out = nn.Linear(dim_heads, dim)
 
         self.bucket_size = bucket_size
         self.lsh_attn = LSHAttention(bucket_size=bucket_size, n_hashes=n_hashes, causal=causal, random_rotations_per_head=random_rotations_per_head, attend_across_buckets = attend_across_buckets,  allow_duplicate_attention = allow_duplicate_attention, return_attn = return_attn, dropout = dropout, **kwargs)
@@ -580,16 +515,16 @@ class LSHSelfAttention(nn.Module):
         self.mem_kv = nn.Parameter(torch.randn(1, num_mem_kv, dim, requires_grad=True)) if num_mem_kv > 0 else None
 
         self.n_local_attn_heads = n_local_attn_heads
-        self.local_attn = LocalAttention(bucket_size=bucket_size * 2, causal=causal, dropout=dropout, shared_qk=True, look_forward=(1 if not causal else 0))
+        self.local_attn = LocalAttention(window_size=bucket_size * 2, causal=causal, dropout=dropout, shared_qk=True, look_forward=(1 if not causal else 0))
 
         self.callback = None
 
     def forward(self, x, keys = None, input_mask = None, input_attn_mask = None, context_mask = None, **kwargs):
         device, dtype = x.device, x.dtype
-        b, t, e, h, m, l_h = *x.shape, self.heads, self.num_mem_kv, self.n_local_attn_heads
+        b, t, e, h, dh, m, l_h = *x.shape, self.heads, self.dim_head, self.num_mem_kv, self.n_local_attn_heads
 
         mem_kv = default(self.mem_kv, torch.empty(b, 0, e, dtype=dtype, device=device))
-        mem = mem_kv.expand(b, m, e)
+        mem = mem_kv.expand(-1, m, -1)
 
         keys = default(keys, torch.empty(b, 0, e, dtype=dtype, device=device))
         c = keys.shape[1]
@@ -640,7 +575,7 @@ class LSHSelfAttention(nn.Module):
         out, attn, buckets = attn_fn_in_chunks(qk, v, **masks)
 
         if self.callback is not None:
-            self.callback(attn.reshape(b, h, t, -1), buckets.reshape(b, h, -1))
+            self.callback(attn.reshape(b, lsh_h, t, -1), buckets.reshape(b, lsh_h, -1))
 
         if has_local:
             lqk, lv = lqk[:, :t], lv[:, :t]
@@ -649,7 +584,7 @@ class LSHSelfAttention(nn.Module):
             out = out.reshape(b, lsh_h, t, -1)
             out = torch.cat((local_out, out), dim=1)
 
-        out = split_heads(out).view(b, t, e)
+        out = split_heads(out).view(b, t, -1)
         out = self.to_out(out)
         return self.post_attn_dropout(out)
 
@@ -702,48 +637,15 @@ class FixedPositionalEmbedding(nn.Module):
         self.register_buffer('inv_freq', inv_freq)
 
     def forward(self, x):
-        t = torch.arange(x.shape[1], device=x.device).type(self.inv_freq.type())
+        t = torch.arange(x.shape[1], device=x.device).type_as(self.inv_freq)
         sinusoid_inp = torch.einsum("i,j->ij", t, self.inv_freq)
         emb = torch.cat((sinusoid_inp.sin(), sinusoid_inp.cos()), dim=-1)
         return emb[None, :, :]
 
-# rewritten from https://github.com/google/trax/blob/master/trax/layers/attention.py#L126
-# with help from @AranKomat
-class AxialPositionalEncoding(nn.Module):
-    def __init__(self, dim, max_seq_len, axial_shape = (), axial_emb_dims = ()):
-        super().__init__()
-        assert sum(axial_emb_dims) == dim, 'axial position embedding dimensions must sum to model dimension'
-        assert reduce(mul, axial_shape, 1) == max_seq_len, 'axial position shape must multiply up to max sequence length'
-
-        self.seq_len = max_seq_len
-        self.shape = axial_shape
-        self.emb_dims = axial_emb_dims
-
-        self.weights = nn.ParameterList([])
-        for ind, (d_emb, shape) in enumerate(zip(self.emb_dims, self.shape)):
-            ax_shape = [1] * len(self.shape)
-            ax_shape[ind] = shape
-            ax_shape = (1, *ax_shape, d_emb)
-            ax_emb = nn.Parameter(torch.zeros(ax_shape).normal_(0, 1))
-            self.weights.append(ax_emb)
-
-    def forward(self, x):
-        b, t, e = x.shape
-        embs = []
-
-        for ax_emb in self.weights:
-            ax_emb_dim = ax_emb.shape[-1]
-            expand_shape = (b, *self.shape, ax_emb_dim)
-            emb = ax_emb.expand(expand_shape).reshape(b, self.seq_len, ax_emb_dim)
-            embs.append(emb)
-
-        pos_emb = torch.cat(embs, dim=-1)
-        return pos_emb[:, :t]
-
 # reformer lm
 
 class Reformer(nn.Module):
-    def __init__(self, dim, depth, max_seq_len, heads = 8, bucket_size = 64, n_hashes = 8, ff_chunks = 100, attn_chunks = None, causal = False, weight_tie = False, lsh_dropout = 0., ff_dropout = 0., ff_activation = None, ff_mult = 4, ff_glu = False, post_attn_dropout = 0., layer_dropout = 0., lsh_attend_across_buckets = True, lsh_allow_duplicate_attention = True, random_rotations_per_head = False, twin_attention = False, use_scale_norm = False, use_rezero = False, use_full_attn = False, full_attn_thres = 0, reverse_thres = 0, num_mem_kv = 0, one_value_head = False, n_local_attn_heads = 0):
+    def __init__(self, dim, depth, max_seq_len, heads = 8, dim_head = None, bucket_size = 64, n_hashes = 8, ff_chunks = 100, attn_chunks = None, causal = False, weight_tie = False, lsh_dropout = 0., ff_dropout = 0., ff_activation = None, ff_mult = 4, ff_glu = False, post_attn_dropout = 0., layer_dropout = 0., lsh_attend_across_buckets = True, lsh_allow_duplicate_attention = True, random_rotations_per_head = False, twin_attention = False, use_scale_norm = False, use_rezero = False, use_full_attn = False, full_attn_thres = 0, reverse_thres = 0, num_mem_kv = 0, one_value_head = False, n_local_attn_heads = 0, pkm_layers = tuple(), pkm_num_keys = 128):
         super().__init__()
         self.dim = dim
         self.depth = depth
@@ -754,12 +656,12 @@ class Reformer(nn.Module):
         self.twin_attention = twin_attention
         self.full_attn_thres = full_attn_thres
 
-        get_attn = lambda: LSHSelfAttention(dim, heads, bucket_size, n_hashes, causal = causal, dropout = lsh_dropout, post_attn_dropout = post_attn_dropout, attn_chunks = attn_chunks, allow_duplicate_attention = lsh_allow_duplicate_attention, attend_across_buckets = lsh_attend_across_buckets, random_rotations_per_head = random_rotations_per_head, num_mem_kv = num_mem_kv, use_full_attn = use_full_attn, full_attn_thres = full_attn_thres, one_value_head = one_value_head, n_local_attn_heads = n_local_attn_heads)
+        get_attn = lambda: LSHSelfAttention(dim, heads, bucket_size, n_hashes, causal = causal, dim_head = dim_head, dropout = lsh_dropout, post_attn_dropout = post_attn_dropout, attn_chunks = attn_chunks, allow_duplicate_attention = lsh_allow_duplicate_attention, attend_across_buckets = lsh_attend_across_buckets, random_rotations_per_head = random_rotations_per_head, num_mem_kv = num_mem_kv, use_full_attn = use_full_attn, full_attn_thres = full_attn_thres, one_value_head = one_value_head, n_local_attn_heads = n_local_attn_heads)
         get_ff = lambda: Chunk(ff_chunks, FeedForward(dim, dropout = ff_dropout, activation = ff_activation, mult = ff_mult, glu = ff_glu), along_dim = -2)
+        get_pkm = lambda: PKM(dim, num_keys = pkm_num_keys)
 
         if weight_tie:
-            get_attn = cache_fn(get_attn)
-            get_ff = cache_fn(get_ff)
+            get_attn, get_ff, get_pkm = map(cache_fn, (get_attn, get_ff, get_pkm))
 
         blocks = []
 
@@ -767,9 +669,19 @@ class Reformer(nn.Module):
 
         residual_fn_wrapper = ReZero if use_rezero else partial(PreNorm, norm_type, dim)
 
-        for _ in range(depth):
+        for ind in range(depth):
+            layer_num = ind + 1
+            use_pkm = layer_num in cast_tuple(pkm_layers)
+            parallel_net = None
+
             attn = get_attn()
-            parallel_net = get_attn() if twin_attention else get_ff()
+
+            if use_pkm:
+                parallel_net = get_pkm()
+            elif twin_attention:
+                parallel_net = get_attn()
+            else:
+                parallel_net = get_ff()
 
             f = residual_fn_wrapper(attn)
             g = residual_fn_wrapper(parallel_net)
@@ -785,7 +697,7 @@ class Reformer(nn.Module):
         return torch.stack(x.chunk(2, dim=-1)).mean(dim=0)
 
 class ReformerLM(nn.Module):
-    def __init__(self, num_tokens, dim, depth, max_seq_len, heads = 8, bucket_size = 64, n_hashes = 4, ff_chunks = 100, attn_chunks = 1, causal = False, weight_tie = False, lsh_dropout = 0., ff_dropout = 0., ff_mult = 4, ff_activation = None, ff_glu = False, post_attn_dropout = 0., layer_dropout = 0., random_rotations_per_head = False, twin_attention = False, use_scale_norm = False, use_rezero = False, use_full_attn = False, full_attn_thres = 0, reverse_thres = 0, num_mem_kv = 0, one_value_head = False, emb_dim = None, return_embeddings = False, weight_tie_embedding = False, fixed_position_emb = False, absolute_position_emb = False, axial_position_shape = None, axial_position_dims = None, n_local_attn_heads = 0):
+    def __init__(self, num_tokens, dim, depth, max_seq_len, heads = 8, dim_head = None, bucket_size = 64, n_hashes = 4, ff_chunks = 100, attn_chunks = 1, causal = False, weight_tie = False, lsh_dropout = 0., ff_dropout = 0., ff_mult = 4, ff_activation = None, ff_glu = False, post_attn_dropout = 0., layer_dropout = 0., random_rotations_per_head = False, twin_attention = False, use_scale_norm = False, use_rezero = False, use_full_attn = False, full_attn_thres = 0, reverse_thres = 0, num_mem_kv = 0, one_value_head = False, emb_dim = None, return_embeddings = False, weight_tie_embedding = False, fixed_position_emb = False, absolute_position_emb = False, axial_position_shape = None, n_local_attn_heads = 0, pkm_layers = tuple(), pkm_num_keys = 128):
         super().__init__()
         emb_dim = default(emb_dim, dim)
         self.max_seq_len = max_seq_len
@@ -800,10 +712,9 @@ class ReformerLM(nn.Module):
             self.pos_emb = FixedPositionalEmbedding(emb_dim)
         else:
             axial_position_shape = default(axial_position_shape, (max_seq_len // bucket_size, bucket_size))
-            axial_position_dims = default(axial_position_dims, (emb_dim // 2, emb_dim // 2))
-            self.pos_emb = AxialPositionalEncoding(emb_dim, max_seq_len, axial_position_shape, axial_position_dims)
+            self.pos_emb = AxialPositionalEmbedding(emb_dim, axial_position_shape)
 
-        self.reformer = Reformer(dim, depth, max_seq_len, heads = heads, bucket_size = bucket_size, n_hashes = n_hashes, ff_chunks = ff_chunks, attn_chunks = attn_chunks, causal = causal, weight_tie = weight_tie, lsh_dropout = lsh_dropout, ff_mult = ff_mult, ff_activation = ff_activation, ff_glu = ff_glu, ff_dropout = ff_dropout, post_attn_dropout = 0., layer_dropout = layer_dropout, random_rotations_per_head = random_rotations_per_head, twin_attention = twin_attention, use_scale_norm = use_scale_norm, use_rezero = use_rezero, use_full_attn = use_full_attn, full_attn_thres = full_attn_thres, reverse_thres = reverse_thres, num_mem_kv = num_mem_kv, one_value_head = one_value_head, n_local_attn_heads = n_local_attn_heads)
+        self.reformer = Reformer(dim, depth, max_seq_len, heads = heads, dim_head = dim_head, bucket_size = bucket_size, n_hashes = n_hashes, ff_chunks = ff_chunks, attn_chunks = attn_chunks, causal = causal, weight_tie = weight_tie, lsh_dropout = lsh_dropout, ff_mult = ff_mult, ff_activation = ff_activation, ff_glu = ff_glu, ff_dropout = ff_dropout, post_attn_dropout = 0., layer_dropout = layer_dropout, random_rotations_per_head = random_rotations_per_head, twin_attention = twin_attention, use_scale_norm = use_scale_norm, use_rezero = use_rezero, use_full_attn = use_full_attn, full_attn_thres = full_attn_thres, reverse_thres = reverse_thres, num_mem_kv = num_mem_kv, one_value_head = one_value_head, n_local_attn_heads = n_local_attn_heads, pkm_layers = pkm_layers, pkm_num_keys = pkm_num_keys)
 
         if return_embeddings:
             self.out = Identity()
@@ -816,7 +727,7 @@ class ReformerLM(nn.Module):
 
     def forward(self, x, **kwargs):
         x = self.token_emb(x)
-        x = x + self.pos_emb(x).type(x.type())
+        x = x + self.pos_emb(x).type_as(x)
 
         x = self.to_model_dim(x)
         x = self.reformer(x, **kwargs)
